/**********************************************
 Deamon fr NDE - stellt die Daten des Rechners
 zusammen.
 
 Copyright (C) 2001 Helmut Fahrion
 
 This file is released under the GPL.
**********************************************/

// Mini ToDo
// Verbose Levels einbauen
// Datei open und close minimieren
// Einzulesende Daten minimieren
// usage schreiben



// alles 
#include "deamon.h"

static char version[] = "0.0.1";	// Ausgabe Versionsnummer

// Globales Zeugs 
static char *programm;
static char *pid_file;		// File in which to keep pid
static char *file_name;		// = "/var/state/cpus/%s";      // Formatstring fr das CPU File
static char cpufile[BUF_LEN];
static char hostname[BUF_LEN];

static sig_atomic_t shutdown = 0;
static sig_atomic_t prune = 0;
static sig_atomic_t reinit = 0;
static sig_atomic_t expire_now = 0;


FILE *d;			// Output File

/////////  Die eigentliche Arbeit !!! ////////////

FILE *
fileopen (char *file)
{
  FILE *ret;

  ret = fopen (file, "r");
  if (!ret)
    {
      syslog (LOG_ERR, "error to open file = %s, err=%s",
	      file, strerror (errno));
      shutdown = 1;
    }

  return ret;
}

void
fileclose (FILE * d)
{
  fclose (d);
}


void
getlocaldata (unsigned long *memory, int *load, int *mips)
{
  char buf[BUF_LEN];
  // Dateien fr Dateneingabe
  FILE *memd, *cpud, *mipsd;

  unsigned long a, b, c, d, e, f;
  float fa, fb, fc;


  // Files werden immer neu geffnet, da sonst nur gecachetes Zeugs
  // gelesen wird.

  memd = fileopen ("/proc/meminfo");
  // Speicher lesen
  fscanf (memd, "%s %s %s %s %s %s\n", buf, buf, buf, buf, buf, buf);
  fscanf (memd, "%s\t%lu %lu %lu %lu %lu %lu\n", buf, &a, &b, &c, &d, &e, &f);
  fclose (memd);

  *memory = (a - f) / 1024;

#ifdef DEBUG
  syslog (LOG_ERR, "freemem=%ul MB", *memory);
#endif

  // Load lesen
  cpud = fileopen ("/proc/loadavg");
  //fseek (cpud, 0, SEEK_SET);
  fscanf (cpud, "%f %f %f ", &fa, &fb, &fc);
  fclose (cpud);

  *load = (int) (fa * 100.0);

#ifdef DEBUG
  syslog (LOG_ERR, "load * 100 =%d f=%f", *load, fa);
#endif

  // Mips lesen
  mipsd = fileopen ("/proc/cpuinfo");
  //fseek (mipsd, 0, SEEK_SET);

  while (strcmp (buf, "bogomips"))
    {
      fscanf (mipsd, "%s", buf);
    }
  fscanf (mipsd, "%s", buf);
  fscanf (mipsd, "%f\n", &fb);

  *mips = (int) (fb / (fa + 1.0));	// ergibt free mips !

#ifdef DEBUG
  syslog (LOG_ERR, "mips free =%d  f=%f", *mips, fb);
#endif

  fclose (mipsd);
}


void
handle_cpu ()
{
  char buf[BUF_LEN];
  unsigned long memory;
  int mips, load;

  // Daten ermitteln
  getlocaldata (&memory, &load, &mips);
  fseek (d, 0, SEEK_SET);
  fprintf (d, "FREEMEM=%d\nLOAD=%d\nMIPS=%d\n", memory, load, mips);
  fflush (d);

}

//////////////////////////////////////////////////


void
usage ()
{
  fprintf (stderr, "Usage: %s [options] path \n", programm);
  fprintf (stderr, "sorry: still in work!\n");
}

static void
become_daemon (void)
{
  FILE *pidfp;
  pid_t pid;
  int nullfd;
  static pid_t my_pid,		// Prozessinformationen
    my_pgrp;

  // Don't BUSY any directories unneccessarily
  chdir ("/");

  // Detach from foreground process
  pid = fork ();
  if (pid > 0)
    exit (0);
  else if (pid < 0)
    {
      fprintf (stderr, "%s: Could not detach process\n", programm);
      exit (1);
    }

  // Open syslog
  openlog ("ndosd", LOG_PID, LOG_DAEMON);
  // Don't report LOG_DEBUG messages
  setlogmask (0x7f);

  if (setpgrp ())
    {
      syslog (LOG_CRIT, "setpgrp: %m");
      exit (1);
    }

  // Initialize global data
  my_pid = getpid ();
  my_pgrp = getpgrp ();

  // Redirect all our file descriptors to /dev/null
  if ((nullfd = open ("/dev/null", O_RDWR)) < 0)
    {
      syslog (LOG_CRIT, "cannot open /dev/null: %m");
      exit (1);
    }

  if (dup2 (nullfd, STDIN_FILENO) < 0 ||
      dup2 (nullfd, STDOUT_FILENO) < 0 || dup2 (nullfd, STDERR_FILENO) < 0)
    {
      syslog (LOG_CRIT, "redirecting file descriptors failed: %m");
      exit (1);
    }
  close (nullfd);

  // Write pid file if requested
  if (pid_file)
    {
      if ((pidfp = fopen (pid_file, "wt")))
	{
	  fprintf (pidfp, "%lu\n", (unsigned long) my_pid);
	  fclose (pidfp);
	}
      else
	{
	  syslog (LOG_WARNING, "failed to write pid file %s: %m", pid_file);
	  pid_file = NULL;
	}
    }
}

//////// die Hndler ////////
static void
sig_shutdown (int sig)
{
  shutdown = sig;
  syslog (LOG_INFO, "Short is the live of a deamon!");

  // Dateien zumachen
  fileclose (d);

  // Datei lschen!
  unlink (file_name);
}

static void
sig_expire (int sig)
{
  expire_now = 1;
}

static void
sig_prune (int sig)
{
  prune = 1;
}

static void
sig_reinit (int sig)
{
  syslog (LOG_INFO, "sig_reinit recieved");

  // force the work
  // handle_cpu ();
  // durch dieses Signal fliegt er aus dem sleep raus, 
  // 
  prune = 1;
  reinit = 1;
}

void
init_signale ()
{
  static struct sigaction sa;	// Signalbehandlung

  // Alles abklemmen oder behandeln !!!
  // The following signals cause a shutdown event to occur
  sa.sa_handler = sig_shutdown;
  sigemptyset (&sa.sa_mask);
  sigaddset (&sa.sa_mask, SIGCHLD);
  sigaddset (&sa.sa_mask, SIGALRM);
  sa.sa_flags = 0;
  sigaction (SIGIO, &sa, NULL);
  sigaction (SIGTERM, &sa, NULL);
  // USR2 is special, but is handled by the same handler
  sigaction (SIGUSR2, &sa, NULL);

  /* The following signals cause a shutdown event to occur, but if we get
     more than one, permit the signal to proceed so we don't loop.  This
     is basically the complete list of "this shouldn't happen" signals. */
  sa.sa_flags = SA_ONESHOT;
  sigaction (SIGXCPU, &sa, NULL);
  sigaction (SIGXFSZ, &sa, NULL);
  sigaction (SIGINT, &sa, NULL);
#ifdef SIGPWR
  sigaction (SIGPWR, &sa, NULL);
#endif
#ifndef DEBUG
  /* When debugging, these signals should be in the default state; when
     in production, we want to at least attempt to catch them and shut down. */
  sigaction (SIGILL, &sa, NULL);
  sigaction (SIGQUIT, &sa, NULL);
  sigaction (SIGABRT, &sa, NULL);
  sigaction (SIGTRAP, &sa, NULL);
  sigaction (SIGFPE, &sa, NULL);
  sigaction (SIGSEGV, &sa, NULL);
  sigaction (SIGBUS, &sa, NULL);
  sigaction (SIGPROF, &sa, NULL);
  sigaction (SIGPIPE, &sa, NULL);
#ifdef SIGSYS
  sigaction (SIGSYS, &sa, NULL);
#endif
#ifdef SIGSTKFLT
  sigaction (SIGSTKFLT, &sa, NULL);
#endif
#ifdef SIGLOST
  sigaction (SIGLOST, &sa, NULL);
#endif
#ifdef SIGEMT
  sigaction (SIGEMT, &sa, NULL);
#endif
#endif // DEBUG

/*
  sa.sa_handler = sig_child;
  sigemptyset(&sa.sa_mask);
  sigaddset(&sa.sa_mask, SIGALRM);
  sa.sa_flags = SA_NOCLDSTOP;   // Don't need info about stopped children
  sigaction(SIGCHLD, &sa, NULL);
*/

  // The SIGALRM handler controls expiration of entries.
  sa.sa_handler = sig_expire;
  sigemptyset (&sa.sa_mask);
  sa.sa_flags = 0;
  sigaction (SIGALRM, &sa, NULL);

  // SIGUSR1 causes a prune event
  sa.sa_handler = sig_prune;
  sigaction (SIGUSR1, &sa, NULL);

  // SIGHUP causes complete reinitialization
  sa.sa_handler = sig_reinit;
  sigaction (SIGHUP, &sa, NULL);

  // The following signals shouldn't occur, and are ignored
  sa.sa_handler = SIG_IGN;
  sigaction (SIGVTALRM, &sa, NULL);
  sigaction (SIGURG, &sa, NULL);
  sigaction (SIGWINCH, &sa, NULL);
#ifdef SIGUNUSED
  sigaction (SIGUNUSED, &sa, NULL);
#endif
}


// Main, bearbeite die Parameter
int
main (int argc, char *argv[])
{
  int opt;

  static const struct option long_options[] = {
    {"help", 0, 0, 'h'},
    {"pid-file", 1, 0, 'p'},
    {"version", 0, 0, 'v'},
    {"file", 1, 0, 'c'},
    {0, 0, 0, 0}
  };

  // Progname frs Logging
  programm = argv[0];

  pid_file = "/var/run/nded.pid";
  file_name = "/var/state/cpus";

  opterr = 0;
  while ((opt =
	  getopt_long (argc, argv, "+hp:f:v", long_options, NULL)) != EOF)
    {
      switch (opt)
	{
	case 'h':
	  usage ();
	  exit (0);
	case 'f':
	  file_name = strdup (optarg);
	  break;
	case 'p':
	  pid_file = strdup (optarg);
	  break;
	case 'v':
	  printf ("%s\n", version);
	  exit (0);
	case '?':
	case ':':
	default:
	  printf ("%s: Ambiguous or unknown options\n", programm);
	  exit (1);
	}
    }

  /*
     // Remove the options
     argv += optind;
     argc -= optind;

     // Pfad angegeben ?
     if (argc < 1)
     {
     usage ();
     exit (1);
     }
   */

  // was ein Deamon noch so machen tut
  become_daemon ();

  syslog (LOG_INFO, "starting %s version %s", programm, version);

  init_signale ();

  /* Initialization successful.  If we're a submount, send outself SIGSTOP to let
     our parent know that we have grown up and don't need supervision anymore. */

  // ### noch testen
  // ##### kill(my_pid, SIGSTOP);


  // generiere Filename
  gethostname (hostname, BUF_LEN);
  if (strlen (hostname) + strlen (file_name) + 1 < BUF_LEN)
    sprintf (cpufile, "%s/%s", file_name, hostname);
  else
    {
      syslog (LOG_ERR,
	      "%s: strlen(hostname) + strlen(file_name) >= BUF_LEN !!!",
	      programm);
      sprintf (cpufile, "%s/%s", file_name, "ndos-hostunknown");
    }

  d = fopen (cpufile, "w+");
  if (!d)
    {
      syslog (LOG_ERR, "error to open file = %s, err=%s",
	      cpufile, strerror (errno));
      shutdown = 1;
    }

  while (!shutdown)
    {
      handle_cpu ();
      usleep (60 * 1000 * 1000);
    }

  fileclose (d);

  syslog (LOG_INFO, "Short is the live of a deamon: %s!", programm);

  return 0;
}
